/* Langton's Ant in C and X11
 * Author: Aaron Weeden, Shodor, 2015
 */
#include <stdbool.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <X11/Xlib.h>

#define BLACK 0
#define WHITE 1

#define UP    0
#define LEFT  1
#define RIGHT 2
#define DOWN  3

// Declare constants
int const STEPS_PER_SIM    = 200;
int const PX_PER_CELL_X    = 32;
int const PX_PER_CELL_Y    = 32;
int const PX_PER_ANT_X     = 16;
int const PX_PER_ANT_Y     = 16;
int const CELLS_PER_GRID_X = 11;
int const CELLS_PER_GRID_Y = 11;
int const MICROS_PER_SLEEP = 5000;

void drawAndPause(Display * const x11Display, Window const x11Window,
                  GC const x11GC,
                  int cellStates[CELLS_PER_GRID_X][CELLS_PER_GRID_Y],
                  XColor const x11WhiteColor, XColor const x11BlackColor,
                  XColor const x11RedColor, int const antDir,
                  int const antCellX, int const antCellY) {
  // Draw cells
  XClearWindow(x11Display, x11Window);
  for (int i = 0; i < CELLS_PER_GRID_X; i++) {
    for (int j = 0; j < CELLS_PER_GRID_Y; j++) {
      XSetForeground(x11Display,
                     x11GC,
                     (cellStates[i][j] == BLACK) ? x11BlackColor.pixel :
                                                   x11WhiteColor.pixel);
      XFillRectangle(x11Display,
                     x11Window,
                     x11GC,
                     i * PX_PER_CELL_X,
                     j * PX_PER_CELL_Y,
                     PX_PER_CELL_X,
                     PX_PER_CELL_Y);
    }
  }

  // Draw ant
  XSetForeground(x11Display,
                 x11GC,
                 x11RedColor.pixel);
  int antPxX;
  int antPxY;
  switch (antDir) {
    case UP:
      antPxX = antCellX * PX_PER_CELL_X +
               0.5 * (PX_PER_CELL_X - PX_PER_ANT_X);
      antPxY = antCellY * PX_PER_CELL_Y;
      break;
    case LEFT:
      antPxX = antCellX * PX_PER_CELL_X;
      antPxY = antCellY * PX_PER_CELL_Y +
               0.5 * (PX_PER_CELL_Y - PX_PER_ANT_Y);
      break;
    case RIGHT:
      antPxX = (antCellX + 1) * PX_PER_CELL_X - PX_PER_ANT_X;
      antPxY = antCellY * PX_PER_CELL_Y +
               0.5 * (PX_PER_CELL_Y - PX_PER_ANT_Y);
      break;
    case DOWN:
      antPxX = antCellX * PX_PER_CELL_X +
               0.5 * (PX_PER_CELL_X - PX_PER_ANT_X);
      antPxY = (antCellY + 1) * PX_PER_CELL_Y - PX_PER_ANT_Y;
      break;
  }
  XFillRectangle(x11Display,
                 x11Window,
                 x11GC,
                 antPxX,
                 antPxY,
                 PX_PER_ANT_X,
                 PX_PER_ANT_Y);
  XFlush(x11Display);

  // Pause
  usleep(MICROS_PER_SLEEP);
}

int main() {
  // Initialize the X Windows Environment
  // This all comes from
  //   http://en.wikibooks.org/wiki/X_Window_Programming/XLib
  //   http://tronche.com/gui/x/xlib-tutorial
  //   http://user.xmission.com/~georgeps/documentation/tutorials/
  //      Xlib_Beginner.html
  Display * const x11Display = XOpenDisplay(NULL);
  if(x11Display == NULL) {
    fprintf(stderr, "ERROR: could not open X display\n");
    exit(EXIT_FAILURE);
  }
  int const x11Screen = DefaultScreen(x11Display);
  Window const x11Window = XCreateSimpleWindow(x11Display,
                                         RootWindow(x11Display, x11Screen),
                                         0,
                                         0,
                                         PX_PER_CELL_X  * CELLS_PER_GRID_X,
                                         PX_PER_CELL_Y  * CELLS_PER_GRID_Y,
                                         1,
                                         BlackPixel(x11Display, x11Screen),
                                         WhitePixel(x11Display, x11Screen));
  Atom deleteWindow = XInternAtom(x11Display, "WM_DELETE_WINDOW", 0);
  XSetWMProtocols(x11Display, x11Window, &deleteWindow, 1);
  XSelectInput(x11Display, x11Window, ExposureMask | KeyPressMask);
  XMapWindow(x11Display, x11Window);
  Colormap const x11ColorMap = DefaultColormap(x11Display, 0);
  GC const x11GC = XCreateGC(x11Display, x11Window, 0, 0);
  XColor x11WhiteColor;
  XColor x11BlackColor;
  XColor x11RedColor;
  XParseColor(x11Display, x11ColorMap, "#000000", &x11BlackColor);
  XParseColor(x11Display, x11ColorMap, "#FFFFFF", &x11WhiteColor);
  XParseColor(x11Display, x11ColorMap, "#FF0000", &x11RedColor);
  XAllocColor(x11Display, x11ColorMap, &x11BlackColor);
  XAllocColor(x11Display, x11ColorMap, &x11WhiteColor);
  XAllocColor(x11Display, x11ColorMap, &x11RedColor);

  // Initialize cell states
  int cellStates[CELLS_PER_GRID_X][CELLS_PER_GRID_Y];
  for (int i = 0; i < CELLS_PER_GRID_X; i++) {
    for (int j = 0; j < CELLS_PER_GRID_Y; j++) {
      cellStates[i][j] = WHITE;
    }
  }

  // Center the ant
  int antCellX = CELLS_PER_GRID_X / 2;
  int antCellY = CELLS_PER_GRID_Y / 2;

  // Give the ant direction
  int antDir = LEFT;

  // Draw/pause
  drawAndPause(x11Display, x11Window, x11GC, cellStates, x11WhiteColor,
               x11BlackColor, x11RedColor, antDir, antCellX, antCellY);

  // Begin animating
  int t = 0;
  while (true) {
    // Only advance the simulation up to the specified number of time steps
    if (t < STEPS_PER_SIM) {
      // Turn
      switch (antDir) {
        case UP:
          antDir = (cellStates[antCellX][antCellY] == BLACK) ? LEFT  : RIGHT;
          break;
        case LEFT:
          antDir = (cellStates[antCellX][antCellY] == BLACK) ? DOWN  : UP;
          break;
        case RIGHT:
          antDir = (cellStates[antCellX][antCellY] == BLACK) ? UP    : DOWN;
          break;
        case DOWN:
          antDir = (cellStates[antCellX][antCellY] == BLACK) ? RIGHT : LEFT;
          break;
      }
  
      drawAndPause(x11Display, x11Window, x11GC, cellStates, x11WhiteColor,
                   x11BlackColor, x11RedColor, antDir, antCellX, antCellY);
  
      // Update cell
      cellStates[antCellX][antCellY] =
        (cellStates[antCellX][antCellY] == BLACK) ? WHITE : BLACK;
  
      drawAndPause(x11Display, x11Window, x11GC, cellStates, x11WhiteColor,
                   x11BlackColor, x11RedColor, antDir, antCellX, antCellY);
  
      // Move
      switch (antDir) {
        case UP:
          antCellY--;
          break;
        case LEFT:
          antCellX--;
          break;
        case RIGHT:
          antCellX++;
          break;
        case DOWN:
          antCellY++;
          break;
      }
  
      drawAndPause(x11Display, x11Window, x11GC, cellStates, x11WhiteColor,
                   x11BlackColor, x11RedColor, antDir, antCellX, antCellY);

      t++;
    } // Stop checking for maximum time steps
  } // Continue animation loop

  return 0;
}

